// Gutierrez, Cesareo
// cgutie@essex.ac.uk
// CC484 Assignment.December 2002

import java.util.*;

/**
 * This class implements the algorithm Learn No Goods studied in the CC484
 * module. Three methods have been implemented: LNCL is the recursive one,
 * analyseNoGood and knownToBeNoGood. These methods implement the three
 * methods descripted in the book Foundations of Constraint Satisfaction
 * by Edward Tsang page 144-145.
 *
 * @version 1.0 10/12/2002
 * @author Cesareo Gutierrez
 */
public class NoGoods {

  /** It has been implemented in a flexible way for the number of queens **/
  static int NUMBERQUEENS = 8;

  /** The set of no good queen combinations found **/
  Vector m_noGood;

  /** The GUI. This class attribute provides access to print the output **/
  QueensGUI m_gui;

  /**
  * Constructor
  * it sets up the initial values.
  */
  public NoGoods(QueensGUI gui) {
    m_noGood = new Vector();
    m_gui = gui;
  }

  /**
  * learnNoGoodAlgorithm
  * The Learn No Good Algorithm. It uses the recursive function.
  *
  * @see NoGoods#LNCL
  *
  * @params Z  Variables
  * @params D  Domains
  * @params C  Constraints
  */
  public void learnNoGoodAlgorithm(MyArray Z, MyArray[] D, Constraints C){
    LNCL(Z,new CompoundLabel(), D, C);
  }

  /**
  * learnNoGoodAlgorithm
  * The Learn No Good Algorithm. It finds a solution of a generic constraint
  * satisfaction problem.
  *
  * @params Z              Variables unlabelled
  * @params compound_label Values already assigned to variables
  * @params D              Domains
  * @params C              Constraints
  * @return                a solution or null
  */
  public CompoundLabel LNCL(MyArray unlabelled, CompoundLabel compound_label,
                            MyArray[] D, Constraints C){
    //If there is no more variables we have found a solution.
    if(unlabelled.isEmpty()){
      m_gui.drawSolution(compound_label);
      return compound_label;
    }
    else{
      //Take a variable unlabelled yet
      int x = unlabelled.pickValue();
      MyArray TD = (MyArray)D[x].clone();

      //For each value of the domain
      while(!TD.isEmpty()){
        int v = TD.pickValue();
        //It assigns this value to the variable, the compound label <x,v>
        CompoundLabel tmp_comp_label = (CompoundLabel)compound_label.clone();
        tmp_comp_label.addCompoundLabel(x,v);
        //It shows the state of the algorithm in the normal console and GUI output
        System.out.println(tmp_comp_label.toString());
        m_gui.drawQueen(x,v);
        //It checks the constraints and if it is known to be not good
        if (C.checkConstraints(tmp_comp_label) && !knownToBeNoGood(tmp_comp_label)){
          //It does the recursive call
          CompoundLabel solution = this.LNCL((MyArray)unlabelled.clone(),
              tmp_comp_label, D, C);
          //if (solution != null) return solution;
        }
        //It updates the GUI
        m_gui.removeQueen(x,v);
      }//end while

      //Try to find information useful to prune the search space
      analyseNoGood(x, compound_label, (MyArray)D[x].clone(), C);
      return null;
    }
  }

  /**
  * analyseNoGood
  * Try to find information useful to prune the search space. It calculates
  * the minimum covering sets over the conflict set and it adds them to the
  * NoGood set in case they have not already been put before or there are
  * not others that are a subset of them.
  *
  * @params x              This variable has just failed.
  * @params compound_label The compound_label created until now.
  * @params Dx             The domain of x
  * @params C              The constraints
  */
  public void analyseNoGood(int x, CompoundLabel compound_label, MyArray Dx,
                            Constraints C){
    //It calculates the conflict set. It uses CompoundLabel.findOutConflictSet method
    Vector[] conflictSet = new Vector[Dx.length];
    for(int i=0; i < Dx.length; i++){
      int a = Dx.pickValue();
      conflictSet[i] = compound_label.findoutConflictSet(x, a, C);
      if(conflictSet[i].isEmpty()){
        return;
      }
    }
    //It calculates the minimum covering sets.
    m_noGood.add(compound_label);
    Vector coveringSet = compound_label.identifyMinimalCoveringSet(conflictSet);
    //It tries to add them to the NoGood set
    for (Iterator i = coveringSet.iterator(); i.hasNext(); ) {
      CompoundLabel item = (CompoundLabel) i.next();
      //It adds it only if it is not already known to be no good.
      if(!knownToBeNoGood(item)){
        //It displays it on the GUI
        m_gui.drawNoGood(item);
        m_noGood.add(item);
      }
    }
  }

  /**
  * knownToBeNoGood
  * For each no good set already learned checks if it is a subset of cl. That
  * would mean that cl is not minimum, then useless.
  *
  * @params cl  the compound label to check
  * @return     if it is already known to be no good or not.
  */
  public boolean knownToBeNoGood(CompoundLabel cl){
    //For each no good set already learned...
    for (Iterator i = m_noGood.iterator(); i.hasNext(); ) {
      CompoundLabel item = (CompoundLabel)i.next();
      //...It checks if it is a subset of cl. That would mean that cl is not
      //minimum, so it is useless.
      if(item.isSubset(cl)){
        return true;
      }
    }
    return false;
  }
}